@@ -2,10 +2,14 @@ package Shell::Config::Generate;
 
 use strict;
 use warnings;
+use 5.008001;
 use Shell::Guess;
+use Carp qw( croak );
+use Exporter ();
 
 # ABSTRACT: Portably generate config for any shell
-our $VERSION = '0.07_01'; # VERSION
+our $VERSION = '0.27_01'; # TRIAL VERSION
+$VERSION = eval $VERSION;
 
 
 sub new
@@ -98,6 +102,14 @@ sub _value_escape_csh
   $value;
 }
 
+sub _value_escape_fish
+{
+  my $value = shift() . '';
+  $value =~ s/([\n])/\\$1/g;
+  $value =~ s/(')/'"$1"'/g;
+  $value;
+}
+
 sub _value_escape_sh
 {
   my $value = shift() . '';
@@ -109,11 +121,55 @@ sub _value_escape_win32
 {
   my $value = shift() . '';
   $value =~ s/%/%%/g;
-  $value =~ s/([&^|<>])/^$1/g;
+  $value =~ s/([&^|<>()])/^$1/g;
   $value =~ s/\n/^\n\n/g;
   $value;
 }
 
+#   `0  Null
+#   `a  Alert bell/beep
+#   `b  Backspace
+#   `f  Form feed (use with printer output)
+#   `n  New line
+#   `r  Carriage return
+# `r`n  Carriage return + New line
+#   `t  Horizontal tab
+#   `v  Vertical tab (use with printer output)
+
+my %ps = ( # microsoft would have to be different
+  "\0" => '`0',
+  "\a" => '`a',
+  "\b" => '`b',
+  "\f" => '`f',
+  "\r" => '`r',
+  "\n" => '`n',
+  "\t" => '`t',
+  #"\v" => '`v',
+);
+
+sub _value_escape_powershell
+{
+  my $value = shift() . '';
+  $value =~ s/(["'`\$#()])/`$1/g;
+  $value =~ s/([\0\a\b\f\r\n\t])/$ps{$1}/eg;
+  $value;
+}
+
+
+sub set_alias
+{
+  my($self, $alias, $command) = @_;
+  
+  push @{ $self->{commands} }, ['alias', $alias, $command]; 
+}
+
+
+sub set_path_sep
+{
+  my($self, $sep) = @_;
+  push @{ $self->{commands} }, ['set_path_sep', $sep];
+}
+
 
 sub generate
 {
@@ -122,6 +178,7 @@ sub generate
   $shell ||= Shell::Guess->running_shell;
   
   my $buffer = '';
+  my $sep    = $shell->is_win32 ? ';' : ':';
 
   if(exists $self->{shebang} && $shell->is_unix)
   {
@@ -140,12 +197,18 @@ sub generate
   {
     my $command = shift @$args;
 
+    if($command eq 'set_path_sep')
+    {
+      $sep = shift @$args;
+      next;
+    }
+    
     # rewrite set_path as set
     if($command eq 'set_path')
     {
       $command = 'set';
       my $name = shift @$args;
-      $args = [$name, join $shell->is_win32 ? ';' : ':', @$args];
+      $args = [$name, join $sep, @$args];
     }
 
     if($command eq 'set')
@@ -156,6 +219,11 @@ sub generate
         $value = _value_escape_csh($value);
         $buffer .= "setenv $name '$value';\n";
       }
+      elsif($shell->is_fish)
+      {
+        $value = _value_escape_fish($value);
+        $buffer .= "set -x $name '$value';\n";
+      }
       elsif($shell->is_bourne)
       {
         $value = _value_escape_sh($value);
@@ -167,9 +235,14 @@ sub generate
         $value = _value_escape_win32($value);
         $buffer .= "set $name=$value\n";
       }
+      elsif($shell->is_power)
+      {
+        $value = _value_escape_powershell($value);
+        $buffer .= "\$env:$name = \"$value\"\n";
+      }
       else
       {
-        die 'don\'t know how to "set" with ' . $shell->name;
+        croak 'don\'t know how to "set" with ' . $shell->name;
       }
     }
 
@@ -178,45 +251,67 @@ sub generate
       my($name, @values) = @$args;
       if($shell->is_c)
       {
-        my $value = join ':', map { _value_escape_csh($_) } @values;
+        my $value = join $sep, map { _value_escape_csh($_) } @values;
         $buffer .= "test \"\$?$name\" = 0 && setenv $name '$value' || ";
         if($command eq 'prepend_path')
-        { $buffer .= "setenv $name '$value':\"\$$name\"" }
+        { $buffer .= "setenv $name '$value$sep'\"\$$name\"" }
         else
-        { $buffer .= "setenv $name \"\$$name\":'$value'" }
+        { $buffer .= "setenv $name \"\$$name\"'$sep$value'" }
         $buffer .= ";\n";
       }
       elsif($shell->is_bourne)
       {
-        my $value = join ':', map { _value_escape_sh($_) } @values;
+        my $value = join $sep, map { _value_escape_sh($_) } @values;
         $buffer .= "if [ -n \"\$$name\" ] ; then\n";
         if($command eq 'prepend_path')
-        { $buffer .= "  $name='$value':\$$name;\n  export $name;\n" }
+        { $buffer .= "  $name='$value$sep'\$$name;\n  export $name;\n" }
         else
-        { $buffer .= "  $name=\$$name:'$value';\n  export $name\n" }
+        { $buffer .= "  $name=\$$name'$sep$value';\n  export $name\n" }
         $buffer .= "else\n";
         $buffer .= "  $name='$value';\n  export $name;\n";
         $buffer .= "fi;\n";
       }
-      elsif($shell->is_cmd || $shell->is_command)
+      elsif($shell->is_fish)
       {
-        my $value = join ';', map { _value_escape_win32($_) } @values;
-        $buffer .= "if defined $name (set ";
+        my $value = join ' ', map { _value_escape_fish($_) } @values;
+        $buffer .= "if [ \"\$$name\" == \"\" ]; set -x $name $value; else; ";
         if($command eq 'prepend_path')
-        { $buffer .= "$name=$value;%$name%" }
+        { $buffer .= "set -x $name $value \$$name;" }
         else
-        { $buffer .= "$name=%$name%;$value" }
-        $buffer .=") else (set $name=$value)\n";
+        { $buffer .= "set -x $name \$$name $value;" }
+        $buffer .= "end\n";
+      }
+      elsif($shell->is_cmd || $shell->is_command || $shell->is_power)
+      {
+        my $value = join $sep, map { $shell->is_power ? _value_escape_powershell($_) : _value_escape_win32($_) } @values;
+        if($shell->is_power)
+        {
+          $buffer .= "if(\$env:$name) { ";
+          if($command eq 'prepend_path')
+          { $buffer .= "\$env:$name = \"$value$sep\" + \$env:$name" }
+          else
+          { $buffer .= "\$env:$name = \$env:$name + \"$sep$value\"" }
+          $buffer .= " } else { \$env:$name = \"$value\" }\n";
+        }
+        else
+        {
+          $buffer .= "if defined $name (set ";
+          if($command eq 'prepend_path')
+          { $buffer .= "$name=$value$sep%$name%" }
+          else
+          { $buffer .= "$name=%$name%$sep$value" }
+          $buffer .=") else (set $name=$value)\n";
+        }
       }
       else
       {
-        die 'don\'t know how to "append_path" with ' . $shell->name;
+        croak 'don\'t know how to "append_path" with ' . $shell->name;
       }
     }
 
     elsif($command eq 'comment')
     {
-      if($shell->is_unix)
+      if($shell->is_unix || $shell->is_power)
       {
         $buffer .= "# $_\n" for map { split /\n/, } @$args;
       }
@@ -226,7 +321,35 @@ sub generate
       }
       else
       {
-        die 'don\'t know how to "comment" with ' . $shell->name;
+        croak 'don\'t know how to "comment" with ' . $shell->name;
+      }
+    }
+    
+    elsif($command eq 'alias')
+    {
+      if($shell->is_bourne)
+      {
+        $buffer .= "alias $args->[0]=\"$args->[1]\";\n";
+      }
+      elsif($shell->is_c)
+      {
+        $buffer .= "alias $args->[0] $args->[1];\n";
+      }
+      elsif($shell->is_cmd || $shell->is_command)
+      {
+        $buffer .= "DOSKEY $args->[0]=$args->[1] \$*\n";
+      }
+      elsif($shell->is_power)
+      {
+        $buffer .= sprintf("function %s { %s \$args }\n", $args->[0], _value_escape_powershell($args->[1]));
+      }
+      elsif($shell->is_fish)
+      {
+        $buffer .= "alias $args->[0] '$args->[1]';\n";
+      }
+      else
+      {
+        croak 'don\'t know how to "alias" with ' . $shell->name;
       }
     }
   }
@@ -238,14 +361,41 @@ sub generate
 sub generate_file
 {
   my($self, $shell, $filename) = @_;
-  use autodie;
-  open my $fh, '>', $filename;
-  print $fh $self->generate($shell);
-  close $fh;
+  my $fh;
+  open($fh, '>', $filename) or die "cannot open $filename: $!";
+  print $fh $self->generate($shell) or die "cannot write $filename: $!";
+  close $fh or die "error closing $filename: $!";
 }
 
-# TODO alias
-# TODO PowerShell
+*import = \&Exporter::import;
+
+our @EXPORT_OK = qw( win32_space_be_gone cmd_escape_path powershell_escape_path );
+
+
+*_win_to_posix_path = $^O =~ /^(cygwin|msys)$/ ? \&Cygwin::win_to_posix_path : sub { $_[0] };
+*_posix_to_win_path = $^O =~ /^(cygwin|msys)$/ ? \&Cygwin::posix_to_win_path : sub { $_[0] };
+
+sub win32_space_be_gone
+{
+  return @_ if $^O !~ /^(MSWin32|cygwin|msys)$/;
+  map { /\s/ ? _win_to_posix_path(Win32::GetShortPathName(_posix_to_win_path($_))) : $_ } @_;
+}
+
+
+sub cmd_escape_path
+{
+  my $path = shift() . '';
+  $path =~ s/%/%%/g;
+  $path =~ s/([&^|<>])/^$1/g;
+  $path =~ s/\n/^\n\n/g;
+  "\"$path\"";
+}
+
+
+sub powershell_escape_path
+{
+  map { my $p = _value_escape_powershell($_); $p =~ s/ /` /g; $p } @_;
+}
 
 1;
 
@@ -253,13 +403,15 @@ __END__
 
 =pod
 
+=encoding UTF-8
+
 =head1 NAME
 
 Shell::Config::Generate - Portably generate config for any shell
 
 =head1 VERSION
 
-version 0.07_01
+version 0.27_01
 
 =head1 SYNOPSIS
 
@@ -389,7 +541,9 @@ or
 
 =head1 CONSTRUCTOR
 
-=head2 Shell::Config::Generate->new
+=head2 new
+
+ my $config = Shell::Config::Generate->new;
 
 creates an instance of She::Config::Generate.
 
@@ -419,11 +573,15 @@ This may be useful for system administrators that must support
 users that use different shells, with a single configuration
 generation script written in Perl.
 
-=head2 $config-E<gt>set( $name => $value )
+=head2 set
+
+ $config->set( $name => $value );
 
 Set an environment variable.
 
-=head2 $config-E<gt>set_path( $name => @values )
+=head2 set_path
+
+ $config->set_path( $name => @values );
 
 Sets an environment variable which is stored in standard
 'path' format (Like PATH or PERL5LIB).  In UNIX land this 
@@ -435,19 +593,25 @@ you do so you have to determine the correct separator.
 This will replace the existing path value if it already
 exists.
 
-=head2 $config-E<gt>append_path( $name => @values );
+=head2 append_path
+
+ $config->append_path( $name => @values );
 
 Appends to an environment variable which is stored in standard
 'path' format.  This will create a new environment variable if
 it doesn't already exist, or add to an existing value.
 
-=head2 $config-E<gt>prepend_path( $name => @values );
+=head2 prepend_path
+
+ $config->prepend_path( $name => @values );
 
 Prepend to an environment variable which is stored in standard
 'path' format.  This will create a new environment variable if
 it doesn't already exist, or add to an existing value.
 
-=head2 $config-E<gt>comment( $comment )
+=head2 comment
+
+ $config->comment( $comment );
 
 This will generate a comment in the appropriate format.
 
@@ -455,7 +619,10 @@ B<note> that including comments in your configuration may mean
 it will not work with the C<eval> backticks method for importing
 configurations into your shell.
 
-=head2 $config-E<gt>shebang( [ $location ] )
+=head2 shebang
+
+ $config->shebang;
+ $config->shebang($location);
 
 This will generate a shebang at the beginning of the configuration,
 making it appropriate for use as a script.  For non UNIX shells this
@@ -467,32 +634,111 @@ B<note> that the shebang in your configuration may mean
 it will not work with the C<eval> backticks method for importing
 configurations into your shell.
 
-=head2 $config-E<gt>echo_off
+=head2 echo_off
+
+ $config->echo_off;
 
 For DOS/Windows configurations (C<command.com> or C<cmd.exe>), issue this as the
 first line of the config:
 
  @echo off
 
-=head2 $config-E<gt>echo_on
+=head2 echo_on
+
+ $config->echo_on;
 
 Turn off the echo off (that is do not put anything at the beginning of
 the config) for DOS/Windows configurations (C<command.com> or C<cmd.exe>).
 
-=head2 $config-E<gt>generate( [ $shell ] )
+=head2 set_alias
+
+ $config->set_alias( $alias => $command )
+
+Sets the given alias to the given command.
+
+Caveat:
+some older shells do not support aliases, such as
+the original bourne shell.  This module will generate
+aliases for those shells anyway, since /bin/sh may
+actually be a more modern shell that DOES support
+aliases, so do not use this method unless you can be
+reasonable sure that the shell you are generating
+supports aliases.  On Windows, for PowerShell, a simple
+function is used instead of an alias so that arguments
+may be specified.
+
+=head2 set_path_sep
+
+ $config->set_path_sep( $sep );
+
+Use C<$sep> as the path separator instead of the shell
+default path separator (generally C<:> for Unix shells 
+and C<;> for Windows shells).
+
+Not all characters are supported, it is usually best
+to stick with the shell default or to use C<:> or C<;>.
+
+=head2 generate
+
+ my $command_text = $config->generate;
+ my $command_text = $config->generate( $shell );
 
 Generate shell configuration code for the given shell.
 $shell is an instance of L<Shell::Guess>.  If $shell
 is not provided, then this method will use Shell::Guess
 to guess the shell that called your perl script.
 
-=head2 $config-E<gt>generate_file( $shell, $filename )
+=head2 generate_file
+
+ $config->generate_file( $shell, $filename );
 
 Generate shell configuration code for the given shell
 and write it to the given file.  $shell is an instance 
 of L<Shell::Guess>.  If there is an IO error it will throw
 an exception.
 
+=head1 FUNCTIONS
+
+=head2 win32_space_be_gone
+
+ my @new_path_list = win32_space_be_gone( @orig_path_list );
+
+On C<MSWin32> and C<cygwin>:
+
+Given a list of directory paths (or filenames), this will
+return an equivalent list of paths pointing to the same 
+file system objects without spaces.  To do this 
+C<Win32::GetShortPathName()> is used on to find alternative
+path names without spaces.
+
+NOTE that this breaks when Windows is told not to create
+short (C<8+3>) filenames; see L<http://www.perlmonks.org/?node_id=333930>
+for a discussion of this behaviour.
+
+In addition, on just C<Cygwin>:
+
+The input paths are first converted from POSIX to Windows paths
+using C<Cygwin::posix_to_win_path>, and then converted back to
+POSIX paths using C<Cygwin::win_to_posix_path>.
+
+Elsewhere:
+
+Returns the same list passed into it
+
+=head2 cmd_escape_path
+
+ my @new_path_list = cmd_escape_path( @orig_path_list )
+
+Given a list of directory paths (or filenames), this will
+return an equivalent list of paths escaped for cmd.exe and command.com.
+
+=head2 powershell_escape_path
+
+ my @new_path_list = powershell_escape_path( @orig_path_list )
+
+Given a list of directory paths (or filenames), this will
+return an equivalent list of paths escaped for PowerShell.
+
 =head1 CAVEATS
 
 The test suite tests this module's output against the actual
@@ -538,15 +784,30 @@ There are probably more clever or prettier ways to
 append/prepend path environment variables as I am not a shell
 programmer.  Patches welcome.
 
-Only UNIX and Windows are supported so far.  Patches welcome.
+Only UNIX (bourne, bash, csh, ksh, fish and their derivatives) and
+Windows (command.com, cmd.exe and PowerShell) are supported so far.
+
+Fish shell support should be considered a tech preview.  The Fish
+shell itself is somewhat in flux, and thus some tests are skipped
+for the Fish shell since behavior is different for different versions.
+In particular, new lines in environment variables may not work on
+newer versions.
+
+Patches welcome for your favorite shell / operating system.
 
 =head1 AUTHOR
 
-Graham Ollis <plicease@cpan.org>
+Author: Graham Ollis E<lt>plicease@cpan.orgE<gt>
+
+Contributors:
+
+Brad Macpherson (BRAD, brad-mac)
+
+mohawk
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2012 by Graham Ollis.
+This software is copyright (c) 2017 by Graham Ollis.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
